from gpsa.models.gpsa import GPSA
from gpsa.models.vgpsa import VariationalGPSA
from gpsa.util.util import (
    rbf_kernel,
    matern12_kernel,
    matern32_kernel,
    polar_warp,
    get_st_coordinates,
    LossNotDecreasingChecker,
)



import torch
import numpy as np
import torch.nn as nn
from ..util.util import rbf_kernel

device = "cuda" if torch.cuda.is_available() else "cpu"

# Define model
class GPSA(nn.Module):
    """
    Args:
        data_dict (dict): Dictionary of data in the format {"modality": {"spatial_coords": X, "outputs": Y, "n_samples_list": n_samples_list}}
        data_init (bool, optional): Whether to initialize inducing locations with KMeans on data.
        n_spatial_dims (int, optional): Number of spatial dimensions (usually 2 or 3).
        n_noise_variance_params (int, optional): Number of noise variance parameters.
        kernel_func_warp (function, optional): Covariance function for warp GP.
        kernel_func_data (function, optional): Covariance function for output GP.
        mean_function (str, optional): Mean function for warp GP. One of ["identity_fixed", "identity_initialized", or None]. None results in a linear mean function.
        mean_penalty_param (float, optional): Description
        fixed_warp_kernel_variances (None, optional): Description
        fixed_warp_kernel_lengthscales (None, optional): Description
        fixed_data_kernel_lengthscales (None, optional): Description
    """

    def __init__(
        self,
        data_dict,
        data_init=True,
        n_spatial_dims=2,
        n_noise_variance_params=2,
        kernel_func_warp=rbf_kernel,
        kernel_func_data=rbf_kernel,
        mean_function="identity_fixed",
        mean_penalty_param=0.0,
        fixed_warp_kernel_variances=None,
        fixed_warp_kernel_lengthscales=None,
        fixed_data_kernel_lengthscales=None,
    ):
        # Constructor
        super(GPSA, self).__init__()
        self.modality_names = list(data_dict.keys())
        self.n_modalities = len(self.modality_names)
        self.mean_penalty_param = mean_penalty_param

        ## Make sure all modalities have the same number of "views"
        n_views = np.unique(
            np.array(
                [len(data_dict[mod]["n_samples_list"]) for mod in self.modality_names]
            )
        )
        if len(n_views) != 1:
            raise ValueError("Each modality must have the same number of views.")
        self.n_views = n_views[0]

        ## Make sure all modalities have the same domain for the spatial coordinates
        n_spatial_dims = np.unique(
            np.array(
                [
                    data_dict[mod]["spatial_coords"].shape[1]
                    for mod in self.modality_names
                ]
            )
        )
        if len(n_spatial_dims) != 1:
            raise ValueError(
                "Each modality must have the same number of spatial dimensions."
            )
        self.n_spatial_dims = n_spatial_dims[0]

        view_idx, Ns, Ps, n_total = self.create_view_idx_dict(data_dict)
        self.view_idx = view_idx
        self.Ns = Ns
        self.Ps = Ps
        self.n_total = n_total
        # import ipdb; ipdb.set_trace()

        ## Number of kernel parameters:
        ##		- 2 parameters for each view for warp GP (lengthscale and variance)
        ##		- 2 parameters for observation GP (lengthscale and variance)
        self.n_kernel_params = 2 * self.n_views + 2
        self.n_noise_variance_params = n_noise_variance_params
        self.kernel_func_warp = kernel_func_warp
        self.kernel_func_data = kernel_func_data

        ## Parameters
        self.noise_variance = nn.Parameter(
            torch.randn([self.n_noise_variance_params]) - 1
        )
        # self.noise_variance = torch.log(torch.ones(2) * 0.001)

        if fixed_warp_kernel_variances is None:
            # self.warp_kernel_variances = nn.Parameter(
            #     torch.randn([self.n_kernel_params // 2 - 1]) - 1
            # )
            self.warp_kernel_variances = nn.Parameter(
                torch.zeros(self.n_kernel_params // 2 - 1)
            )
        else:
            self.warp_kernel_variances = torch.log(
                torch.tensor(fixed_warp_kernel_variances)
            )

        if fixed_warp_kernel_lengthscales is None:
            # self.warp_kernel_lengthscales = nn.Parameter(
            #     torch.randn([self.n_kernel_params // 2 - 1]) + 3
            # )
            self.warp_kernel_lengthscales = nn.Parameter(
                torch.zeros(self.n_kernel_params // 2 - 1) + np.log(10)
            )
        else:
            self.warp_kernel_lengthscales = torch.log(
                torch.tensor(fixed_warp_kernel_lengthscales)
            )

        if fixed_data_kernel_lengthscales is None:
            self.data_kernel_lengthscale = nn.Parameter(
                torch.log(torch.exp(torch.randn(1)))
            )
        else:
            self.data_kernel_lengthscale = torch.log(
                torch.tensor(fixed_data_kernel_lengthscales).float()
            )

        self.data_kernel_variance = nn.Parameter(torch.randn(1))
        # self.data_kernel_variance = nn.Parameter(torch.zeros(1))
        # self.data_kernel_variance = torch.tensor(0.).float()

        if mean_function == "identity_fixed":
            self.mean_slopes = (
                torch.eye(self.n_spatial_dims, device=device)
                .unsqueeze(0)
                .repeat(self.n_views, 1, 1)
            )
            self.mean_intercepts = torch.zeros(
                [self.n_views, self.n_spatial_dims], device=device
            )
        elif mean_function == "identity_initialized":
            self.mean_slopes = nn.Parameter(
                torch.randn([self.n_views, self.n_spatial_dims, self.n_spatial_dims])
            )
            self.mean_intercepts = nn.Parameter(
                torch.zeros([self.n_views, self.n_spatial_dims])
            )
        else:
            self.mean_slopes = nn.Parameter(
                torch.eye(self.n_spatial_dims).unsqueeze(0).repeat(self.n_views, 1, 1)
            )
            self.mean_intercepts = nn.Parameter(
                torch.randn([self.n_views, self.n_spatial_dims]) * 0.1
            )

        # self.mean_intercepts = torch.zeros([self.n_views, self.n_spatial_dims])
        self.diagonal_offset = 1e-5

    def create_view_idx_dict(self, data_dict):
        """Summary

        Args:
            data_dict (TYPE): Description

        Returns:
            TYPE: Description
        """
        view_idx, Ns, Ps = {}, {}, {}
        n_total = 0
        for mod in self.modality_names:
            n_samples_list = data_dict[mod]["n_samples_list"]
            # self.n_samples_lists[mod] = n_samples_list
            curr_N = np.sum(n_samples_list)
            Ns[mod] = curr_N
            n_total += curr_N
            Ps[mod] = data_dict[mod]["outputs"].shape[1]

            # Compute the indices of each view for each modality
            cumulative_sums = np.cumsum(n_samples_list)
            cumulative_sums = np.insert(cumulative_sums, 0, 0)
            curr_view_idx = [
                np.arange(cumulative_sums[ii], cumulative_sums[ii + 1])
                for ii in range(self.n_views)
            ]
            view_idx[mod] = curr_view_idx

        return view_idx, Ns, Ps, n_total

    def compute_mean_penalty(self):
        return self.mean_penalty_param * torch.mean(
            torch.square(
                self.mean_slopes
                - torch.eye(self.n_spatial_dims).unsqueeze(0).repeat(self.n_views, 1, 1)
            )
        )

    def forward(self, X_spatial):
        raise (NotImplementedError)

    def loss_fn(self, data_dict, Gs, means_G_list, covs_G_list, means_Y, covs_Y):
        raise (NotImplementedError)


def distance_matrix(X, Y):
    """Compute distances between samples (rows) of two matrices

    Args:
        X (array): n x D matrix of spatial locations
        Y (array): m x D matrix of spatial locations

    Returns:
        array: n x m matrix whose ij'th elementh is the Euclidean distance between i'th row of X and j'th row of Y
    """
    squared_diffs = torch.square(torch.unsqueeze(X, 0) - torch.unsqueeze(Y, 1))
    squared_distances = torch.sum(squared_diffs, dim=2)
    return squared_distances


if __name__ == "__main__":

    pass



import torch
import numpy as np
import torch.nn as nn
from sklearn.cluster import KMeans
from gpsa import GPSA
from ..util.util import rbf_kernel
from collections.abc import Iterable

torch.autograd.set_detect_anomaly(True)

device = "cuda" if torch.cuda.is_available() else "cpu"


class VariationalGPSA(GPSA):
    def __init__(
        self,
        data_dict,
        m_X_per_view,
        m_G,
        data_init=True,
        minmax_init=False,
        grid_init=False,
        n_spatial_dims=2,
        n_noise_variance_params=2,
        kernel_func_warp=rbf_kernel,
        kernel_func_data=rbf_kernel,
        n_latent_gps=None,
        mean_function="identity_fixed",
        mean_penalty_param=0.0,
        fixed_warp_kernel_variances=None,
        fixed_warp_kernel_lengthscales=None,
        fixed_data_kernel_lengthscales=None,
        fixed_view_idx=None,
    ):
        super(VariationalGPSA, self).__init__(
            data_dict,
            data_init=True,
            n_spatial_dims=2,
            n_noise_variance_params=2,
            kernel_func_warp=kernel_func_warp,
            kernel_func_data=kernel_func_data,
            mean_penalty_param=mean_penalty_param,
            fixed_warp_kernel_variances=fixed_warp_kernel_variances,
            fixed_warp_kernel_lengthscales=fixed_warp_kernel_lengthscales,
            fixed_data_kernel_lengthscales=fixed_data_kernel_lengthscales,
        )

        self.m_X_per_view = m_X_per_view
        self.m_G = m_G
        self.n_latent_gps = n_latent_gps
        self.n_latent_outputs = {}
        for mod in self.modality_names:
            curr_n_latent_outputs = (
                self.n_latent_gps[mod]
                if self.n_latent_gps[mod] is not None
                else self.Ps[mod]
            )
            self.n_latent_outputs[mod] = curr_n_latent_outputs
        self.fixed_view_idx = fixed_view_idx

        if data_init:
            # Initialize inducing locations with a subset of the data
            Xtilde = torch.zeros([self.n_views, self.m_X_per_view, self.n_spatial_dims])
            for ii in range(self.n_views):
                curr_X_spatial_list = []
                for mod in self.modality_names:
                    curr_idx = self.view_idx[mod][ii]
                    curr_modality_and_view_spatial = data_dict[mod]["spatial_coords"][
                        curr_idx, :
                    ]
                    curr_X_spatial_list.append(curr_modality_and_view_spatial)
                curr_X_spatial = torch.cat(curr_X_spatial_list, dim=0)

                kmeans = KMeans(n_clusters=self.m_X_per_view)
                kmeans.fit(curr_X_spatial.detach().cpu().numpy())
                Xtilde[ii, :, :] = torch.tensor(kmeans.cluster_centers_)

            self.Xtilde = nn.Parameter(Xtilde.clone())
            # self.Xtilde = Xtilde.clone()

            rand_idx = np.random.choice(
                np.arange(curr_X_spatial.shape[0]),
                size=self.m_G,
                replace=False,
            )

            all_X_spatial = torch.cat(
                [data_dict[mod]["spatial_coords"] for mod in self.modality_names]
            )
            kmeans = KMeans(n_clusters=self.m_G)
            kmeans.fit(all_X_spatial.detach().cpu().numpy())
            self.Gtilde = nn.Parameter(torch.tensor(kmeans.cluster_centers_))

        elif grid_init:

            if self.n_spatial_dims == 2:
                xlow, ylow = (
                    data_dict[self.modality_names[0]]["spatial_coords"].numpy().min(0)
                )
                xhigh, yhigh = (
                    data_dict[self.modality_names[0]]["spatial_coords"].numpy().max(0)
                )
                xlimits = [xlow, xhigh]
                ylimits = [ylow, yhigh]
                numticks = np.ceil(np.sqrt(self.m_G)).astype(int)
                self.m_G = numticks**2
                self.m_X_per_view = numticks**2
                x1s = np.linspace(*xlimits, num=numticks)
                x2s = np.linspace(*ylimits, num=numticks)
                X1, X2 = np.meshgrid(x1s, x2s)
                Xtilde = np.vstack([X1.ravel(), X2.ravel()]).T
                Xtilde_torch = torch.zeros(
                    [self.n_views, Xtilde.shape[0], self.n_spatial_dims]
                )
                for vv in range(self.n_views):
                    Xtilde_torch[vv] = torch.tensor(Xtilde)

                # self.Xtilde = Xtilde_torch.clone()
                # self.Gtilde = torch.tensor(Xtilde).float()
                self.Xtilde = nn.Parameter(Xtilde_torch.clone())
                self.Gtilde = nn.Parameter(torch.tensor(Xtilde).float())

        else:
            # Random initialization of inducing locations
            self.Xtilde = nn.Parameter(
                torch.randn([self.n_views, self.m_X_per_view, self.n_spatial_dims])
            )
            self.Gtilde = nn.Parameter(torch.randn([self.m_G, self.n_spatial_dims]))

        ## Variational covariance parameters
        Omega_sqt_G_list = torch.zeros(
            [self.n_views * self.n_spatial_dims, self.m_X_per_view, self.m_X_per_view],
            device=device,
        )
        for ii in range(self.n_views):
            for jj in range(self.n_spatial_dims):
                Omega_sqt = 0.1 * torch.randn(
                    size=[self.m_X_per_view, self.m_X_per_view]
                )
                # import ipdb; ipdb.set_trace()
                # Omega_sqt_G_list[ii * self.n_views + jj, :, :] = Omega_sqt
                Omega_sqt_G_list[jj * self.n_views + ii, :, :] = Omega_sqt
        self.Omega_sqt_G_list = nn.Parameter(Omega_sqt_G_list)

        Omega_sqt_F_dict = torch.nn.ParameterDict()
        for mod in self.modality_names:
            num_outputs = self.Ps[mod]
            curr_Omega = torch.zeros([self.n_latent_outputs[mod], self.m_G, self.m_G])
            for jj in range(self.n_latent_outputs[mod]):
                Omega_sqt = 0.1 * torch.randn(size=[self.m_G, self.m_G])
                curr_Omega[jj, :, :] = Omega_sqt
            Omega_sqt_F_dict[mod] = nn.Parameter(curr_Omega)
        self.Omega_sqt_F_dict = Omega_sqt_F_dict

        ## Variational mean parameters
        self.delta_G_list = nn.Parameter(self.Xtilde.clone())
        delta_F_dict = torch.nn.ParameterDict()
        for mod in self.modality_names:
            num_outputs = self.Ps[mod]
            curr_delta = nn.Parameter(
                torch.randn(size=[self.m_G, self.n_latent_outputs[mod]], device=device)
            )
            delta_F_dict[mod] = curr_delta
        self.delta_F_dict = delta_F_dict

        ## LMC parameters
        self.W_dict = torch.nn.ParameterDict()
        for mod in self.modality_names:
            if self.n_latent_gps[mod] is not None:
                self.W_dict[mod] = nn.Parameter(
                    torch.randn([self.n_latent_gps[mod], self.Ps[mod]])
                )

    def compute_mean_and_var(
        self, Kff_diag, Kuf, Kuu_chol, mu_x, mu_z, delta, Omega_tril
    ):
        alpha_x = torch.cholesky_solve(Kuf, Kuu_chol)

        a_t_Kchol = torch.matmul(alpha_x.transpose(-1, -2), Kuu_chol)
        aKa = torch.sum(torch.square(a_t_Kchol), dim=-1)

        mu_tilde = mu_x.unsqueeze(0) + torch.matmul(
            alpha_x.transpose(-1, -2), delta - mu_z
        )

        if len(alpha_x.shape) == 2:
            a_t_Omega_tril = torch.matmul(
                alpha_x.transpose(-1, -2).unsqueeze(0), Omega_tril
            )
            aOmega_a = torch.sum(torch.square(a_t_Omega_tril), dim=-1)
            Sigma_tilde = Kff_diag - aKa + aOmega_a + self.diagonal_offset
        else:
            a_t_Omega_tril = torch.matmul(
                alpha_x.transpose(-1, -2).unsqueeze(1), Omega_tril.unsqueeze(0)
            )
            aOmega_a = torch.sum(torch.square(a_t_Omega_tril), dim=-1)
            Sigma_tilde = (
                Kff_diag.unsqueeze(1)
                - aKa.unsqueeze(1)
                + aOmega_a
                + self.diagonal_offset
            )

        return mu_tilde, Sigma_tilde + self.diagonal_offset

    def get_Omega_from_Omega_sqt(self, Omega_sqt):
        return torch.matmul(
            Omega_sqt,
            torch.transpose(Omega_sqt, -1, -2),
        ) + self.diagonal_offset * torch.eye(Omega_sqt.shape[-1], device=device)

    def forward(self, X_spatial, view_idx, Ns, S=1, prediction_mode=False, G_test=None):

        if prediction_mode:
            self.eval()

        self.noise_variance_pos = torch.exp(self.noise_variance) + self.diagonal_offset

        self.mu_z_G = (
            torch.zeros(
                [self.n_views, self.m_X_per_view, self.n_spatial_dims], device=device
            )
            * np.nan
        )
        for vv in range(self.n_views):
            self.mu_z_G[vv] = (
                torch.mm(self.Xtilde[vv], self.mean_slopes[vv])
                + self.mean_intercepts[vv]
            )
            if self.fixed_view_idx is not None and (
                vv in self.fixed_view_idx
                if isinstance(self.fixed_view_idx, Iterable)
                else self.fixed_view_idx == vv
            ):
                self.mu_z_G[vv] *= 100.0

        self.Kuu_chol_list = (
            torch.zeros(
                [self.n_views, self.m_X_per_view, self.m_X_per_view], device=device
            )
            * np.nan
        )
        G_samples = {}
        for mod in self.modality_names:
            G_samples[mod] = (
                torch.zeros([S, Ns[mod], self.n_spatial_dims], device=device) * np.nan
            )

        G_means = {}
        for mod in self.modality_names:
            G_means[mod] = (
                torch.zeros([Ns[mod], self.n_spatial_dims], device=device) * np.nan
            )

        curr_Omega_G = self.get_Omega_from_Omega_sqt(self.Omega_sqt_G_list)

        self.curr_Omega_tril_list = torch.cholesky(curr_Omega_G)

        for vv in range(self.n_views):

            ## If this view is fixed (template-based alignment), then we don't need to sample for it.
            if self.fixed_view_idx is not None and (
                vv in self.fixed_view_idx
                if isinstance(self.fixed_view_idx, Iterable)
                else self.fixed_view_idx == vv
            ):
                for mm, mod in enumerate(self.modality_names):
                    observed_X_spatial = X_spatial[mod][view_idx[mod][vv]]
                    G_means[mod][view_idx[mod][vv]] = observed_X_spatial

                    G_samples[mod][:, view_idx[mod][vv], :] = observed_X_spatial

                continue

            kernel_G = lambda x1, x2, diag=False: self.kernel_func_warp(
                x1,
                x2,
                lengthscale_unconstrained=self.warp_kernel_lengthscales[vv],
                output_variance_unconstrained=self.warp_kernel_variances[vv],
                diag=diag,
            )

            ## Collect data from all modalities for this view
            curr_X_spatial_list = []
            curr_n = 0
            curr_mod_idx = []
            for mod in self.modality_names:
                curr_idx = view_idx[mod][vv]
                curr_mod_idx.append(np.arange(curr_n, curr_n + len(curr_idx)))
                curr_n += len(curr_idx)
                curr_modality_and_view_spatial = X_spatial[mod][curr_idx, :]
                curr_X_spatial_list.append(curr_modality_and_view_spatial)

            curr_X_spatial = torch.cat(curr_X_spatial_list, dim=0)

            if len(curr_X_spatial) == 0:
                continue

            curr_X_tilde = self.Xtilde[vv]

            mu_x_G = (
                torch.mm(curr_X_spatial, self.mean_slopes[vv])
                + self.mean_intercepts[vv]
            )

            # Kff_diag = (
            #     kernel_G(curr_X_spatial, curr_X_spatial, diag=True)
            #     + self.diagonal_offset
            # )
            Kff_diag = torch.ones((curr_X_spatial.shape[0]), device=device) * torch.exp(
                self.warp_kernel_variances[vv]
            )

            Kuu = kernel_G(
                curr_X_tilde, curr_X_tilde
            ) + self.diagonal_offset * torch.eye(self.m_X_per_view, device=device)

            Kuf = kernel_G(curr_X_tilde, curr_X_spatial)

            Kuu_chol = torch.cholesky(Kuu)
            self.Kuu_chol_list[vv, :, :] = Kuu_chol

            mu_tilde, Sigma_tilde = self.compute_mean_and_var(
                Kff_diag,
                Kuf,
                Kuu_chol,
                mu_x_G,
                self.mu_z_G,
                self.delta_G_list,
                self.curr_Omega_tril_list,
            )

            # Sample
            G_marginal_dist = torch.distributions.Normal(
                mu_tilde[vv],
                Sigma_tilde[
                    vv * self.n_spatial_dims : vv * self.n_spatial_dims
                    + self.n_spatial_dims
                ].t(),
            )

            for mm, mod in enumerate(self.modality_names):
                curr_idx = curr_mod_idx[mm]
                G_means[mod][view_idx[mod][vv]] = mu_tilde[vv][curr_idx]

            for ss in range(S):

                curr_G_sample = G_marginal_dist.rsample()
                for mm, mod in enumerate(self.modality_names):
                    curr_idx = curr_mod_idx[mm]
                    G_samples[mod][ss, view_idx[mod][vv]] = curr_G_sample[curr_idx]

        self.curr_Omega_tril_F = {}
        for mod in self.modality_names:
            self.curr_Omega_tril_F[mod] = torch.zeros(
                [self.n_latent_outputs[mod], self.m_G, self.m_G], device=device
            )

        F_samples = {}
        self.F_latent_samples = {}
        self.F_observed_samples = {}
        for mod in self.modality_names:
            F_samples[mod] = torch.zeros([S, Ns[mod], self.n_latent_outputs[mod]])
            self.F_latent_samples[mod] = torch.zeros(
                [S, Ns[mod], self.n_latent_outputs[mod]], device=device
            )
            self.F_observed_samples[mod] = torch.zeros([S, Ns[mod], self.Ps[mod]])

        if G_test is not None:

            self.F_latent_samples_test = {}
            self.F_observed_samples_test = {}
            for mod in self.modality_names:
                n_test = G_test[mod].shape[1]
                self.F_latent_samples_test[mod] = torch.zeros(
                    [S, n_test, self.n_latent_outputs[mod]]
                )
                self.F_observed_samples_test[mod] = torch.zeros(
                    [S, n_test, self.Ps[mod]]
                )

        kernel_F = lambda x1, x2, diag=False: self.kernel_func_data(
            x1,
            x2,
            lengthscale_unconstrained=self.data_kernel_lengthscale,
            output_variance_unconstrained=self.data_kernel_variance,
            diag=diag,
        )

        Kuu = kernel_F(self.Gtilde, self.Gtilde) + self.diagonal_offset * torch.eye(
            self.m_G, device=device
        )

        self.Kuu_chol_F = torch.cholesky(Kuu)

        for mod in self.modality_names:

            mu_x_F = torch.zeros([Ns[mod], self.n_latent_outputs[mod]], device=device)
            mu_z_F = torch.zeros([self.m_G, self.n_latent_outputs[mod]], device=device)

            # Kff_diag = (
            #     kernel_F(G_samples[mod], G_samples[mod], diag=True)
            #     + self.diagonal_offset
            # )
            Kff_diag = torch.ones(
                (G_samples[mod].shape[:2]), device=device
            ) * torch.exp(self.data_kernel_variance)

            Kuf = kernel_F(self.Gtilde, G_samples[mod])
            curr_Omega = self.get_Omega_from_Omega_sqt(self.Omega_sqt_F_dict[mod])

            self.curr_Omega_tril_F[mod] = torch.cholesky(curr_Omega)
            mu_tilde, Sigma_tilde = self.compute_mean_and_var(
                Kff_diag,
                Kuf,
                self.Kuu_chol_F,
                mu_x_F,
                mu_z_F,
                self.delta_F_dict[mod],
                self.curr_Omega_tril_F[mod],
            )

            eps = torch.randn(mu_tilde.shape, device=device)
            curr_F_latent_samples = (
                mu_tilde + torch.sqrt(torch.transpose(Sigma_tilde, 1, 2)) * eps
            )

            if self.n_latent_gps[mod] is not None:
                curr_W = self.W_dict[mod]
                F_observed_mean = torch.matmul(curr_F_latent_samples, curr_W)
            else:
                F_observed_mean = curr_F_latent_samples

            self.F_latent_samples[mod] = curr_F_latent_samples
            self.F_observed_samples[mod] = F_observed_mean

            ## For test samples
            if G_test is not None:
                # Kff_diag = (
                #     kernel_F(G_samples[mod], G_samples[mod], diag=True)
                #     + self.diagonal_offset
                # )
                Kff_diag = torch.ones(
                    (G_test[mod].shape[:2]), device=device
                ) * torch.exp(
                    self.data_kernel_variance,
                )

                mu_x_F = torch.zeros(
                    [G_test[mod].shape[1], self.n_latent_outputs[mod]], device=device
                )

                Kuf = kernel_F(self.Gtilde, G_test[mod])

                mu_tilde, Sigma_tilde = self.compute_mean_and_var(
                    Kff_diag,
                    Kuf,
                    self.Kuu_chol_F,
                    mu_x_F,
                    mu_z_F,
                    self.delta_F_dict[mod],
                    self.curr_Omega_tril_F[mod],
                )

                eps = torch.randn(mu_tilde.shape, device=device)
                curr_F_latent_samples = (
                    mu_tilde + torch.sqrt(torch.transpose(Sigma_tilde, 1, 2)) * eps
                )

                if self.n_latent_gps[mod] is not None:
                    curr_W = self.W_dict[mod]
                    F_observed_mean = torch.matmul(curr_F_latent_samples, curr_W)
                else:
                    F_observed_mean = curr_F_latent_samples

                self.F_latent_samples_test[mod] = curr_F_latent_samples
                self.F_observed_samples_test[mod] = F_observed_mean

        if G_test is not None:
            return (
                G_means,
                G_samples,
                self.F_latent_samples,
                self.F_observed_samples,
                self.F_latent_samples_test,
                self.F_observed_samples_test,
            )
        else:
            return G_means, G_samples, self.F_latent_samples, self.F_observed_samples

    def loss_fn(self, data_dict, F_samples):
        # This computes the the negative (approximate) ELBO

        # Running sum for KL terms
        KL_div = 0

        ## G
        for vv in range(self.n_views):
            if self.fixed_view_idx is not None and (
                vv in self.fixed_view_idx
                if isinstance(self.fixed_view_idx, Iterable)
                else self.fixed_view_idx == vv
            ):
                continue
            for jj in range(self.n_spatial_dims):
                qu = torch.distributions.MultivariateNormal(
                    loc=self.delta_G_list[vv, :, jj],
                    scale_tril=self.curr_Omega_tril_list[jj * self.n_views + vv, :, :],
                )
                pu = torch.distributions.MultivariateNormal(
                    loc=self.mu_z_G[vv, :, jj],
                    scale_tril=self.Kuu_chol_list[vv, :, :],
                )
                curr_KL_div = torch.distributions.kl.kl_divergence(qu, pu)

                KL_div += curr_KL_div

        ## F
        LL = 0
        pu = torch.distributions.MultivariateNormal(
            loc=torch.zeros(self.m_G, device=device), scale_tril=self.Kuu_chol_F
        )
        for mm, mod in enumerate(self.modality_names):
            qu = torch.distributions.MultivariateNormal(
                loc=self.delta_F_dict[mod].t(),
                scale_tril=self.curr_Omega_tril_F[mod],
            )

            curr_KL_div = torch.distributions.kl.kl_divergence(qu, pu)
            KL_div += curr_KL_div.sum()

            Y_distribution = torch.distributions.Normal(
                loc=F_samples[mod],
                scale=self.noise_variance_pos[-self.n_modalities + mm],
            )
            S = F_samples[mod].shape[0]

            LL += Y_distribution.log_prob(data_dict[mod]["outputs"]).sum() / S

        return -LL + KL_div


if __name__ == "__main__":
    pass



import numpy as np
import pandas as pd
import numpy.random as npr
import torch
from scipy.special import xlogy


def rbf_kernel(
    x1, x2, lengthscale_unconstrained, output_variance_unconstrained, diag=False
):

    lengthscale = torch.exp(lengthscale_unconstrained)
    output_variance = torch.exp(output_variance_unconstrained)

    if diag:
        diffs = x1 - x2
    else:
        diffs = x1.unsqueeze(-2) - x2.unsqueeze(-3)

    K = output_variance * torch.exp(
        -0.5 * torch.sum(torch.square(diffs / lengthscale), dim=-1)
    )
    return K


def rbf_kernel_numpy(x, xp, kernel_params):
    output_scale = np.exp(kernel_params[0])
    lengthscales = np.exp(kernel_params[1:])
    diffs = np.expand_dims(x / lengthscales, 1) - np.expand_dims(xp / lengthscales, 0)
    return output_scale * np.exp(-0.5 * np.sum(diffs**2, axis=2))


def matern12_kernel(
    x1, x2, lengthscale_unconstrained, output_variance_unconstrained, diag=False
):

    lengthscale = torch.exp(lengthscale_unconstrained)
    output_variance = torch.exp(output_variance_unconstrained)

    if diag:
        diffs = x1 - x2
    else:
        diffs = x1.unsqueeze(-2) - x2.unsqueeze(-3)
    eps = 1e-10
    dists = torch.sqrt(torch.sum(torch.square(diffs), dim=-1) + eps)

    return output_variance * torch.exp(-0.5 * dists / lengthscale)


def matern32_kernel(
    x1, x2, lengthscale_unconstrained, output_variance_unconstrained, diag=False
):

    lengthscale = torch.exp(lengthscale_unconstrained)
    output_variance = torch.exp(output_variance_unconstrained)

    if diag:
        diffs = x1 - x2
    else:
        diffs = x1.unsqueeze(-2) - x2.unsqueeze(-3)
    eps = 1e-10
    dists = torch.sqrt(torch.sum(torch.square(diffs), dim=-1) + eps)

    inner_term = np.sqrt(3.0) * dists / lengthscale
    K = output_variance * (1 + inner_term) * torch.exp(-inner_term)
    return K


def polar_warp(X, r, theta):
    return np.array([X[:, 0] + r * np.cos(theta), X[:, 1] + r * np.sin(theta)]).T


def get_st_coordinates(df):
    """
    Extracts spatial coordinates from ST data with index in 'AxB' type format.

    Return: pandas dataframe of coordinates
    """
    coor = []
    for spot in df.index:
        coordinates = spot.split("x")
        coordinates = [float(i) for i in coordinates]
        coor.append(coordinates)
    return np.array(coor)


def compute_distance(X1, X2):
    return np.mean(np.sqrt(np.sum((X1 - X2) ** 2, axis=1)))


def make_pinwheel(
    radial_std, tangential_std, num_classes, num_per_class, rate, rs=npr.RandomState(0)
):
    """Based on code by Ryan P. Adams."""
    rads = np.linspace(0, 2 * np.pi, num_classes, endpoint=False)

    features = rs.randn(num_classes * num_per_class, 2) * np.array(
        [radial_std, tangential_std]
    )
    features[:, 0] += 1
    labels = np.repeat(np.arange(num_classes), num_per_class)

    angles = rads[labels] + rate * np.exp(features[:, 0])
    rotations = np.stack(
        [np.cos(angles), -np.sin(angles), np.sin(angles), np.cos(angles)]
    )
    rotations = np.reshape(rotations.T, (-1, 2, 2))

    return np.einsum("ti,tij->tj", features, rotations)


class ConvergenceChecker(object):
    def __init__(self, span, dtp="float64"):
        self.span = span
        x = np.arange(span, dtype=dtp)
        x -= x.mean()
        X = np.column_stack((np.ones(shape=x.shape), x, x**2, x**3))
        self.U = np.linalg.svd(X, full_matrices=False)[0]

    def smooth(self, y):
        return self.U @ (self.U.T @ y)

    def subset(self, y, idx=-1):
        span = self.U.shape[0]
        lo = idx - span + 1
        if idx == -1:
            return y[lo:]
        else:
            return y[lo : (idx + 1)]

    def relative_change(self, y, idx=-1, smooth=True):
        y = self.subset(y, idx=idx)
        if smooth:
            y = self.smooth(y)
        prev = y[-2]
        return (y[-1] - prev) / (0.1 + abs(prev))

    def converged(self, y, tol=1e-4, **kwargs):
        return abs(self.relative_change(y, **kwargs)) < tol

    def relative_change_all(self, y, smooth=True):
        n = len(y)
        span = self.U.shape[0]
        cc = np.tile([np.nan], n)
        for i in range(span, n):
            cc[i] = self.relative_change(y, idx=i, smooth=smooth)
        return cc

    def converged_all(self, y, tol=1e-4, smooth=True):
        cc = self.relative_change_all(y, smooth=smooth)
        return np.abs(cc) < tol


# Function for computing size factors
def compute_size_factors(m):
    # given matrix m with samples in the columns
    # compute size factors

    sz = np.sum(m.values, axis=0)  # column sums (sum of counts in each cell)
    lsz = np.log(sz)

    # make geometric mean of sz be 1 for poisson
    sz_poisson = np.exp(lsz - np.mean(lsz))
    return sz_poisson


def poisson_deviance(X, sz):

    LP = X.values / sz  # recycling
    # import ipdb; ipdb.set_trace()
    LP[LP > 0] = np.log(LP[LP > 0])  # log transform nonzero elements only

    # Transpose to make features in cols, observations in rows
    X = X.T
    ll_sat = np.sum(np.multiply(X, LP.T), axis=0)
    feature_sums = np.sum(X, axis=0)
    ll_null = feature_sums * np.log(feature_sums / np.sum(sz))
    return 2 * (ll_sat - ll_null)


def deviance_feature_selection(X):

    # Remove cells without any counts
    X = X[np.sum(X, axis=1) > 0]

    # Compute size factors
    sz = compute_size_factors(X)

    # Compute deviances
    devs = poisson_deviance(X, sz)

    # Get associated gene names
    gene_names = X.index.values

    assert gene_names.shape[0] == devs.values.shape[0]

    return devs.values, gene_names


def deviance_residuals(x, theta, mu=None):
    """Computes deviance residuals for NB model with a fixed theta"""

    if mu is None:
        counts_sum0 = np.sum(x, axis=0, keepdims=True)
        counts_sum1 = np.sum(x, axis=1, keepdims=True)
        counts_sum = np.sum(x)
        # get residuals
        mu = counts_sum1 @ counts_sum0 / counts_sum

    def remove_negatives(sqrt_term):
        negatives_idx = sqrt_term < 0
        if np.any(negatives_idx):
            n_negatives = np.sum(negatives_idx)
            print(
                "Setting %u negative sqrt term values to 0 (%f%%)"
                % (n_negatives, n_negatives / np.product(sqrt_term.shape))
            )
            sqrt_term[negatives_idx] = 0

    if np.isinf(theta):  ### POISSON
        x_minus_mu = x - mu
        sqrt_term = 2 * (
            xlogy(x, x / mu) - x_minus_mu
        )  # xlogy(x,x/mu) computes xlog(x/mu) and returns 0 if x=0
        remove_negatives(sqrt_term)
        dev = np.sign(x_minus_mu) * np.sqrt(sqrt_term)
    else:  ### NEG BIN
        x_plus_theta = x + theta
        sqrt_term = 2 * (
            xlogy(x, x / mu) - (x_plus_theta) * np.log(x_plus_theta / (mu + theta))
        )  # xlogy(x,x/mu) computes xlog(x/mu) and returns 0 if x=0
        remove_negatives(sqrt_term)
        dev = np.sign(x - mu) * np.sqrt(sqrt_term)

    return dev


def pearson_residuals(counts, theta, clipping=True):
    """Computes analytical residuals for NB model with a fixed theta, clipping outlier residuals to sqrt(N)"""
    counts_sum0 = np.sum(counts, axis=0, keepdims=True)
    counts_sum1 = np.sum(counts, axis=1, keepdims=True)
    counts_sum = np.sum(counts)

    # get residuals
    mu = counts_sum1 @ counts_sum0 / counts_sum
    z = (counts - mu) / np.sqrt(mu + mu**2 / theta)

    # clip to sqrt(n)
    if clipping:
        n = counts.shape[0]
        z[z > np.sqrt(n)] = np.sqrt(n)
        z[z < -np.sqrt(n)] = -np.sqrt(n)

    return z


class LossNotDecreasingChecker:
    def __init__(self, max_epochs, atol=1e-2, window_size=10):
        self.max_epochs = max_epochs
        self.atol = atol
        self.window_size = window_size
        self.decrease_in_loss = np.zeros(max_epochs)
        self.average_decrease_in_loss = np.zeros(max_epochs)

    def check_loss(self, iternum, loss_trace):

        if iternum >= 1:
            self.decrease_in_loss[iternum] = (
                loss_trace[iternum - 1] - loss_trace[iternum]
            )
            if iternum >= self.window_size:
                self.average_decrease_in_loss[iternum] = np.mean(
                    self.decrease_in_loss[iternum - self.window_size + 1 : iternum]
                )
                has_converged = self.average_decrease_in_loss[iternum] < self.atol
                return has_converged

        return False



import torch
import numpy as np
import matplotlib.pyplot as plt
from torch.utils.data import Dataset, DataLoader
import time
import pandas as pd
from scipy.stats import pearsonr

from matplotlib.lines import Line2D

import seaborn as sns

SCATTER_POINT_SIZE = 50

device = "cuda" if torch.cuda.is_available() else "cpu"

def callback_oned(
    model,
    X,
    Y,
    X_aligned,
    data_expression_ax,
    latent_expression_ax,
    prediction_ax=None,
    X_test=None,
    Y_pred=None,
    Y_test_true=None,
    X_test_aligned=None,
    F_samples=None,
):
    model.eval()
    markers = list(Line2D.markers.keys())
    colors = ["blue", "orange"]

    if model.fixed_view_idx is not None:
        curr_idx = model.view_idx["expression"][model.fixed_view_idx]
        X_aligned["expression"][curr_idx] = torch.tensor(X[curr_idx].astype(np.float32))

    data_expression_ax.cla()
    latent_expression_ax.cla()

    data_expression_ax.set_title("Observed data")
    latent_expression_ax.set_title("Aligned data")

    data_expression_ax.set_xlabel("Spatial coordinate")
    latent_expression_ax.set_xlabel("Spatial coordinate")

    data_expression_ax.set_ylabel("Outcome")
    latent_expression_ax.set_ylabel("Outcome")

    data_expression_ax.set_xlim([X.min(), X.max()])
    latent_expression_ax.set_xlim([X.min(), X.max()])

    for vv in range(model.n_views):

        view_idx = model.view_idx["expression"]

        data_expression_ax.scatter(
            X[view_idx[vv], 0],
            Y[view_idx[vv], 0],
            label="View {}".format(vv + 1),
            marker=markers[vv],
            s=SCATTER_POINT_SIZE,
            c="blue",
        )
        if Y.shape[1] > 1:
            data_expression_ax.scatter(
                X[view_idx[vv], 0],
                Y[view_idx[vv], 1],
                label="View {}".format(vv + 1),
                marker=markers[vv],
                s=SCATTER_POINT_SIZE,
                c="orange",
            )
        latent_expression_ax.scatter(
            # model.G_means["expression"].detach().cpu().numpy()[view_idx[vv], 0],
            X_aligned["expression"].detach().cpu().numpy()[view_idx[vv], 0],
            Y[view_idx[vv], 0],
            c="blue",
            label="View {}".format(vv + 1),
            marker=markers[vv],
            s=SCATTER_POINT_SIZE,
        )
        if Y.shape[1] > 1:
            latent_expression_ax.scatter(
                # model.G_means["expression"].detach().cpu().numpy()[view_idx[vv], 0],
                X_aligned["expression"].detach().cpu().numpy()[view_idx[vv], 0],
                Y[view_idx[vv], 1],
                c="orange",
                label="View {}".format(vv + 1),
                marker=markers[vv],
                s=SCATTER_POINT_SIZE,
            )
        # latent_expression_ax.scatter(
        # 	model.Xtilde.detach().cpu().numpy()[vv, :, 0],
        # 	model.delta_list.detach().cpu().numpy()[vv][:, 0],
        # 	c="red",
        # 	label="View {}".format(vv + 1),
        # 	marker="^",
        # 	s=100,
        # )

        if F_samples is not None:
            latent_expression_ax.scatter(
                X_aligned["expression"].detach().cpu().numpy()[view_idx[vv], 0],
                F_samples.detach().cpu().numpy()[view_idx[vv], 0],
                c="red",
                marker=markers[vv],
                s=SCATTER_POINT_SIZE,
            )
            if Y.shape[1] > 1:
                latent_expression_ax.scatter(
                    X_aligned["expression"].detach().cpu().numpy()[view_idx[vv], 0],
                    F_samples.detach().cpu().numpy()[view_idx[vv], 1],
                    c="green",
                    marker=markers[vv],
                    s=SCATTER_POINT_SIZE,
                )

    if prediction_ax is not None:

        prediction_ax.cla()
        prediction_ax.set_title("Predictions")
        prediction_ax.set_xlabel("True outcome")
        prediction_ax.set_ylabel("Predicted outcome")

        ### Plots the warping function
        # prediction_ax.scatter(
        # 	X[view_idx[vv], 0],
        # 	X_aligned["expression"].detach().cpu().numpy()[view_idx[vv], 0],
        # 	label="View {}".format(vv + 1),
        # 	marker=markers[vv],
        # 	s=100,
        # 	c="blue",
        # )
        # prediction_ax.scatter(
        # 	model.Xtilde.detach().cpu().numpy()[vv, :, 0],
        # 	model.delta_list.detach().cpu().numpy()[vv][:, 0],
        # 	c="red",
        # 	label="View {}".format(vv + 1),
        # 	marker="^",
        # 	s=100,
        # )
        latent_expression_ax.scatter(
            X_test_aligned["expression"].detach().cpu().numpy()[:, 0],
            Y_pred.detach().cpu().numpy()[:, 0],
            c="blue",
            label="Prediction",
            marker="^",
            s=SCATTER_POINT_SIZE,
        )
        latent_expression_ax.scatter(
            X_test_aligned["expression"].detach().cpu().numpy()[:, 0],
            Y_pred.detach().cpu().numpy()[:, 1],
            c="orange",
            label="Prediction",
            marker="^",
            s=SCATTER_POINT_SIZE,
        )
        prediction_ax.scatter(
            Y_test_true[:, 0],
            Y_pred.detach().cpu().numpy()[:, 0],
            c="black",
            s=SCATTER_POINT_SIZE,
        )
        prediction_ax.scatter(
            Y_test_true[:, 1],
            Y_pred.detach().cpu().numpy()[:, 1],
            c="black",
            s=SCATTER_POINT_SIZE,
            marker="^",
        )

    data_expression_ax.legend()
    plt.draw()
    plt.pause(1 / 60.0)


def callback_twod(
    model,
    X,
    Y,
    X_aligned,
    data_expression_ax,
    latent_expression_ax,
    is_mle=False,
    gene_idx=0,
    s=200,
    include_legend=False,
):

    if model.fixed_view_idx is not None:
        if is_mle:
            pass
        else:
            curr_idx = model.view_idx["expression"][model.fixed_view_idx]
            X_aligned["expression"][curr_idx] = torch.tensor(
                X[curr_idx].astype(np.float32), device=device
            )

    model.eval()
    markers = [".", "+", "^"]
    colors = ["blue", "orange"]

    data_expression_ax.cla()
    latent_expression_ax.cla()
    data_expression_ax.set_title("Observed data")
    latent_expression_ax.set_title("Aligned data")

    curr_view_idx = model.view_idx["expression"]

    latent_Xs = []
    Xs = []
    Ys = []
    markers_list = []
    viewname_list = []

    for vv in range(model.n_views):

        ## Data
        Xs.append(X[curr_view_idx[vv]])

        ## Latents
        curr_latent_Xs = X_aligned["expression"].detach().cpu().numpy()[curr_view_idx[vv]]
        latent_Xs.append(curr_latent_Xs)
        Ys.append(Y[curr_view_idx[vv], gene_idx])
        markers_list.append([markers[vv]] * curr_latent_Xs.shape[0])
        viewname_list.append(
            ["Observation {}".format(vv + 1)] * curr_latent_Xs.shape[0]
        )

    Xs = np.concatenate(Xs, axis=0)
    latent_Xs = np.concatenate(latent_Xs, axis=0)
    Ys = np.concatenate(Ys)
    markers_list = np.concatenate(markers_list)
    viewname_list = np.concatenate(viewname_list)

    data_df = pd.DataFrame(
        {
            "X1": Xs[:, 0],
            "X2": Xs[:, 1],
            "Y": Ys,
            "marker": markers_list,
            "view": viewname_list,
        }
    )

    latent_df = pd.DataFrame(
        {
            "X1": latent_Xs[:, 0],
            "X2": latent_Xs[:, 1],
            "Y": Ys,
            "marker": markers_list,
            "view": viewname_list,
        }
    )

    plt.sca(data_expression_ax)
    g = sns.scatterplot(
        data=data_df,
        x="X1",
        y="X2",
        hue="Y",
        style="view",
        ax=data_expression_ax,
        s=s,
        linewidth=1.8,
        edgecolor="black",
        palette="viridis",
    )
    if not include_legend:
        g.legend_.remove()
    # plt.colorbar()
    # plt.axis("off")
    # plt.scatter(model.Xtilde.detach().cpu().numpy()[0, :, 0], model.Xtilde.detach().cpu().numpy()[0, :, 1], color="red")
    # plt.scatter(model.Xtilde.detach().cpu().numpy()[1, :, 0], model.Xtilde.detach().cpu().numpy()[1, :, 1], color="red")
    # plt.scatter(model.Gtilde.detach().cpu().numpy()[:, 0], model.Gtilde.detach().cpu().numpy()[:, 1], color="red")
    # plt.axis("off")

    plt.sca(latent_expression_ax)
    g = sns.scatterplot(
        data=latent_df,
        x="X1",
        y="X2",
        hue="Y",
        style="view",
        ax=latent_expression_ax,
        s=s,
        linewidth=1.8,
        edgecolor="black",
        palette="viridis",
    )
    if not include_legend:
        g.legend_.remove()
    # plt.colorbar()

    # import ipdb; ipdb.set_trace()

    # for vv in range(model.n_views):

    #     # import ipdb; ipdb.set_trace()
    #     data_expression_ax.scatter(
    #         X[curr_view_idx[vv], 0],
    #         X[curr_view_idx[vv], 1],
    #         c=Y[curr_view_idx[vv], 0],
    #         label="View {}".format(vv + 1),
    #         marker=markers[vv],
    #         s=400,
    #     )
    # latent_expression_ax.scatter(
    #     X_aligned["expression"].detach().cpu().numpy()[curr_view_idx[vv], 0],
    #     X_aligned["expression"].detach().cpu().numpy()[curr_view_idx[vv], 1],
    #     c=Y[curr_view_idx[vv], 0],
    #     label="View {}".format(vv + 1),
    #     marker=markers[vv],
    #     s=400,
    # )
    # plt.axis("off")


def callback_twod_aligned_only(
    model,
    X,
    Y,
    X_aligned,
    latent_expression_ax1,
    latent_expression_ax2,
    is_mle=False,
    gene_idx=0,
):

    if model.fixed_view_idx is not None:
        if is_mle:
            pass
        else:
            curr_idx = model.view_idx["expression"][model.fixed_view_idx]
            X_aligned["expression"][curr_idx] = torch.tensor(
                X[curr_idx].astype(np.float32)
            )

    model.eval()
    markers = [".", "+", "^"]
    colors = ["blue", "orange"]

    latent_expression_ax1.cla()
    latent_expression_ax2.cla()
    latent_expression_ax1.set_title("Observed data")
    latent_expression_ax2.set_title("Aligned data")

    curr_view_idx = model.view_idx["expression"]

    latent_Xs = []
    Xs = []
    Ys = []
    markers_list = []
    viewname_list = []

    aligned_coords = X_aligned["expression"].detach().cpu().numpy()

    for vv in range(model.n_views):

        ## Data
        Xs.append(X[curr_view_idx[vv]])

        ## Latents
        curr_latent_Xs = aligned_coords[curr_view_idx[vv]]
        latent_Xs.append(curr_latent_Xs)
        Ys.append(Y[curr_view_idx[vv], gene_idx])
        markers_list.append([markers[vv]] * curr_latent_Xs.shape[0])
        viewname_list.append(["View {}".format(vv + 1)] * curr_latent_Xs.shape[0])

    latent_expression_ax1.scatter(
        aligned_coords[curr_view_idx[0]][:, 0],
        aligned_coords[curr_view_idx[0]][:, 1],
        c=Y[curr_view_idx[0]][:, gene_idx].squeeze(),
        s=24,
        marker="h",
    )
    latent_expression_ax2.scatter(
        aligned_coords[curr_view_idx[1]][:, 0],
        aligned_coords[curr_view_idx[1]][:, 1],
        c=Y[curr_view_idx[1]][:, gene_idx].squeeze(),
        s=24,
        marker="h",
    )
    # latent_expression_ax1.scatter(model.Xtilde.detach().cpu().numpy()[0, :, 0], model.Xtilde.detach().cpu().numpy()[0, :, 1], color="red")
    # latent_expression_ax2.scatter(model.Xtilde.detach().cpu().numpy()[1, :, 0], model.Xtilde.detach().cpu().numpy()[1, :, 1], color="red")

    plt.axis("off")


def callback_twod_multimodal(
    model, data_dict, X_aligned, axes, rgb=False, scatterpoint_size=100
):

    # if model.fixed_view_idx is not None:
    #     if is_mle:
    #         pass
    #     else:
    #         curr_idx = model.view_idx["expression"][model.fixed_view_idx]
    #         X_aligned["expression"][curr_idx] = torch.tensor(X[curr_idx].astype(np.float32))

    model.eval()
    markers = [".", "+", "^"]
    colors = ["blue", "orange"]

    [ax.cla() for ax in axes]

    axes[0].set_title("Observed expression")
    axes[1].set_title("Aligned expression")
    axes[2].set_title("Observed histology")
    axes[3].set_title("Aligned histology")

    axis_counter = 0
    n_mods = 2
    for mod in ["expression", "histology"]:
        curr_view_idx = model.view_idx[mod]
        for vv in range(model.n_views):

            # import ipdb; ipdb.set_trace()
            curr_coords = data_dict[mod]["spatial_coords"]

            if mod == "histology" and rgb:
                curr_outputs = data_dict[mod]["outputs"][curr_view_idx[vv], :]
            else:
                curr_outputs = data_dict[mod]["outputs"][curr_view_idx[vv], 0]
            axes[axis_counter].scatter(
                curr_coords[curr_view_idx[vv], 0],
                curr_coords[curr_view_idx[vv], 1],
                c=curr_outputs,
                label="View {}".format(vv + 1),
                marker=markers[vv],
                s=scatterpoint_size,
            )
            axes[axis_counter + 1].scatter(
                X_aligned[mod].detach().cpu().numpy()[curr_view_idx[vv], 0],
                X_aligned[mod].detach().cpu().numpy()[curr_view_idx[vv], 1],
                c=curr_outputs,
                label="View {}".format(vv + 1),
                marker=markers[vv],
                s=scatterpoint_size,
            )
        axis_counter += n_mods



# tests/test_variational_gpsa_forward.py

import warnings
warnings.filterwarnings(
    "ignore",
    message="torch.cholesky is deprecated in favor of torch.linalg.cholesky",
)

import numpy as np
import torch
from gpsa import VariationalGPSA
from gpsa import rbf_kernel

def test_variational_gpsa_forward_runs_without_error():
    # 1. Create toy one‐view data
    np.random.seed(0)
    X = np.random.rand(50, 2)
    Y = np.sin(2 * np.pi * X[:, 0])[:, None] + 0.1 * np.random.randn(50, 1)

    x = torch.tensor(X, dtype=torch.float32)
    y = torch.tensor(Y, dtype=torch.float32)
    view_idx       = [np.arange(X.shape[0])]
    n_samples_list = [X.shape[0]]

    data_dict = {
        "demo": {
            "spatial_coords": x,
            "outputs":        y,
            "n_samples_list": n_samples_list,
        }
    }

    # 2. Instantiate the variational model
    model = VariationalGPSA(
        data_dict,
        n_spatial_dims=2,
        m_G=10,
        m_X_per_view=10,
        data_init=True,
        minmax_init=False,
        grid_init=False,
        n_latent_gps={"demo": None},
        mean_function="identity_fixed",
        kernel_func_warp=rbf_kernel,
        kernel_func_data=rbf_kernel,
        fixed_view_idx=0,
    ).to("cpu")

    # 3. One forward pass
    view_dict, Ns, _, _ = model.create_view_idx_dict(data_dict)
    G_means, G_samples, F_latent_samples, F_samples = model.forward(
        {"demo": x},
        view_idx=view_dict,
        Ns=Ns,
        S=5,
    )

    # 4. Basic assertions to ensure shapes look right
    assert "demo" in G_means
    assert G_means["demo"].ndim == 2

    # F_samples is a dict, so ensure it has the expected structure
    assert isinstance(F_samples, dict)
    assert "demo" in F_samples
    demo_tensor = F_samples["demo"]
    assert isinstance(demo_tensor, torch.Tensor)
    assert demo_tensor.ndim >= 2



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="gpsa-engelhardt",
    version="0.6.14",  # bump if 0.6.14 was already uploaded to PyPI
    author="Andy Jones and Barbara Engelhardt",
    author_email="engelhardtgpsa@gmail.com",
    description="Gaussian Process Spatial Alignment (GPSA)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/engelhardtgpsa/spatial-alignment",
    packages=find_packages(
        include=["gpsa", "gpsa.*"],
        exclude=("tests", "tests.*", "gpsa.tests", "gpsa.tests.*"),
    ),  # <-- CLOSES find_packages HERE
    python_requires=">=3.10, <3.12",
    license="Apache-2.0",  # keep consistent with classifier & README
    install_requires=[
        "torch==2.2.2",        # keep pinned if you want “just works” installs
        "numpy==1.26.4",
        "pandas==2.2.2",
        "scikit-learn==1.3.2",
        "pcpca==0.3",
        "plottify==1.1",
        "scanpy==1.9.1",
        "scipy==1.12.0",
        "seaborn>=0.13.2",
        "statsmodels==0.14.0",
        "anndata==0.8.0",
        "auto_mix_prep==0.2.0",
        "matplotlib>=3.8.0",
        "tqdm==4.64.0",
        "squidpy==1.1.2",
    ],
    extras_require={
        "dev": ["flake8", "pytest", "pip-tools"],
        "docs": ["pdoc", "py-cpuinfo", "Cython", "numpy"],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: Apache Software License",
    ],
    include_package_data=True,
)
